#!/usr/bin/env python2.7
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
#from __future__ import unicode_literals
from . import lahmandb
from MySQLdb.cursors import DictCursor, Cursor
import numpy as np
from theshowutil.ratinggenerator import BasicRatingGenerator


def outs2inn(outs):
    outs = int(round(outs))
    return ('{}.{}'.format(outs // 3, outs % 3) if outs % 3 > 0
            else '{}  '.format(outs // 3))


def in2ft(inch):
    inch = int(round(inch))
    return '{:1d}-{:d}'.format(inch // 12, inch % 12)


class NoPlayerFoundError(Exception): pass
class NoPlayingTimeError(Exception): pass


class PlayerData(object):

    def __init__(self, playerid, currentyear, yweight={}, db=lahmandb):
        self.playerid = playerid
        self.currentyear = int(currentyear)

        if yweight is None or len(yweight) == 0:
            y = self.currentyear
            yweight = {y: 2., (y-1): 1., (y-2): 1.}
        wtot = sum([float(yw) for yw in yweight.values()])
        for y in yweight:
            yweight[y] = (yweight[y] / wtot) * len(yweight)
        self.yweight = yweight

        self._db = db
        self._init()

    def _init(self):
        pid = self.playerid

        c = self._db.cursor(DictCursor)

        c.execute("SELECT * FROM Master WHERE playerID = %s", (pid,))
        m = c.fetchall()
        if len(m) != 1:
            s = ("invalid playerID: {:s}".format(pid))
            raise NoPlayerFoundError(s)
        self.master = m[0]

        c = self._db.cursor()

        c.execute("SELECT * FROM Pitching WHERE playerID = %s", (pid,))
        self.pitching = lahmandb.Pitching.fromiter(c.fetchall())

        c.execute("SELECT * FROM Batting WHERE playerID = %s", (pid,))
        self.batting = lahmandb.Batting.fromiter(c.fetchall())

        c.execute("SELECT * FROM Fielding WHERE playerID = %s", (pid,))
        self.fielding = lahmandb.Fielding.fromiter(c.fetchall())

        c.execute("SELECT * FROM Appearances WHERE playerID = %s", (pid,))
        self.appearances = lahmandb.Appearances.fromiter(c.fetchall())

        attr = PlayerAttributes()
        attr.update_basic(self)
        self.attr = attr
        
        self.update_total()

    def age(self, year=None):
        """Player's age in given year (or currentyear)"""
        y, m = self.master['birthYear'], self.master['birthMonth']
        return (year if year else self.currentyear) - y - (m > 6)

    def update_total(self):
        w = self.yweight
        y = self.currentyear
        t = {}
        t['pitching'] = TotPitching.weighted(self.pitching, y, w)
        t['batting'] = TotBatting.weighted(self.batting, y, w)
        t['fielding'] = TotFielding.weighted(self.fielding, y, w)
        t['appearances'] = TotAppearances.weighted(self.appearances, y, w)
        self.total4rating = t

        c = BasicRatingGenerator(self.attr, self.total4rating)
        self.attr = c.rate()

    def generate_csv(self):
        return self.attr.generate_csv()


class PlayerAttributes(dict):

    def __init__(self, *args, **kwargs):
        super(PlayerAttributes, self).__init__(*args, **kwargs)
        self.reset()

    def reset(self):
        attr = {'FNM': '', 'LNM': '', 'HT': '', 'WT': '', 'PO1': '',
                'PO2': '', 'T': '', 'B': '', 'USB': '', 'AGE': -1,

                'STA': -1, 'PCL': -1, 'H9': -1, 'HR9': -1, 'K9': -1, 'BB9': -1,
                'P1': '', 'P1S': -1, 'P1C': -1, 'P1B': -1,
                'P2': '', 'P2S': -1, 'P2C': -1, 'P2B': -1,
                'P3': '', 'P3S': -1, 'P3C': -1, 'P3B': -1,
                'P4': '', 'P4S': -1, 'P4C': -1, 'P4B': -1,
                'P5': '', 'P5S': -1, 'P5C': -1, 'P5B': -1,

                'RCT': -1, 'LCT': -1, 'RPW': -1, 'LPW': -1, 'BNT': -1,
                'DBN': -1, 'VIS': -1, 'DIS': -1, 'CLU': -1, 'DUR': -1,
                'SPD': -1, 'AMS': -1, 'AMA': -1, 'REA': -1, 'FLD': -1,
                'BLK': -1, 'BAB': -1, 'BAG': -1}
        self.update(attr)

    def update_basic(self, playerdata, year=None):
        master = playerdata.master
        self['FNM'] = master['nameFirst']
        self['LNM'] = master['nameLast']
        self['HT'] = in2ft(master['height'])
        self['WT'] = master['weight']
        self['T'] = master['throws']
        self['B'] = {'R': 'R', 'L': 'L', 'B': 'S'}[master['bats']]
        self['USB'] = 'Y' if master['birthCountry'] == 'USA' else 'N'
        self['AGE'] = playerdata.age(year)

    @property
    def csvheader(self):
        h = ("#FULLNAME                  AGE HT    WT POS      T B US"
             "  RCT LCT RPW LPW BNT DBN VIS DIS CLU"
             "  DUR SPD AMS AMA REA FLD BLK BAB BAG"
             "  STA PCL  H9 HR9  K9 BB9"
             "  PT1 VL CT BR"
             "  PT2 VL CT BR"
             "  PT3 VL CT BR"
             "  PT4 VL CT BR"
             "  PT5 VL CT BR")
        return h

    @property
    def csv(self):
        d = self.copy()
        d['FULLNAME'] = '"{LNM:s}, {FNM:s}"'.format(**d)
        d['POS'] = d['PO1'] + (',{:s}'.format(d['PO2']) if d['PO2'] else '')
        if (d['PO1'] in ('SP', 'RP', 'CP')
            or d['PO2'] in ('SP', 'RP', 'CP')):
            fmp = ("{STA:2d}  {PCL:2d}  {H9:2d}  {HR9:2d}  {K9:2d}  {BB9:2d}"
                   " {P1:>4s} {P1S:2d} {P1C:2d} {P1B:2d}"
                   " {P2:>4s} {P2S:2d} {P2C:2d} {P2B:2d}"
                   " {P3:>4s} {P3S:2d} {P3C:2d} {P3B:2d}"
                   " {P4:>4s} {P4S:2d} {P4C:2d} {P4B:2d}"
                   " {P5:>4s} {P5S:2d} {P5C:2d} {P5B:2d}")
            sp = fmp.format(**d)
        else:
            sp = ""
        fmt = ('{FULLNAME:27s} {AGE:2d} {HT:<4s} {WT:3d} {POS:<8}'
               " {T:1s} {B:1s}  {USB:1s}"
               "   {RCT:2d}  {LCT:2d}  {RPW:2d}  {LPW:2d}  {BNT:2d}  {DBN:2d}"
               "  {VIS:2d}  {DIS:2d}  {CLU:2d}   {DUR:2d}  {SPD:2d}"
               "  {AMS:2d}  {AMA:2d}  {REA:2d}  {FLD:2d}  {BLK:2d}"
               "  {BAB:2d}  {BAG:2d}   " + sp)
        return fmt.format(**d)


def total(stats, fields):
    tot = {}
    for field in fields:
        tot[field] = sum(stats[field].filled(0.))
    return tot


def weighted_total(stats, fields, year, yweight=None, samplingmode='optimal'):
    if len(stats) == 0:
        # no stats recorded, so no total
        return {}

    if samplingmode == 'career':
        return total(stats)

    #def cmpkey(s):
    #    return s['yearID']

    # list all yearID values in order of summing.
    #ys_weighted = [s for s in stats if s['yearID'] in yweight.keys()]
    #ys = sorted([s for s in stats if s['yearID'] not in ys_weighted
    #             and y['yearID'] <= year], key=cmpkey, reverse=True)
    #ys.extend(sorted([y for y in self if y['yearID'] not in ys_weighted
    #                  and y['yearID'] > year], key=cmpkey))

    tot = {}
    for y in yweight:
        w = yweight[y]
        s = stats[stats['yearID'] == y]
        if len(s) == 0:
            continue
        for field in fields:
            tot[field] = (tot.setdefault(field, 0.)
                          + sum(s[field].filled(0.)) * w)
        tot['YR_TOT'] = tot.setdefault('YR_TOT', 0.) + w

    #if len(tot):
    #    tot['YRS'] = len(set([y for y in stats['yearID'])))

    ## if self.minimum_opportunity is not None:
    ##         aux_weight = (min(yweight.values()) / 2. if len(yweight.values())
    ##                       else 1.)
    ##         nmin, fields = self.minimum_opportunity

    ##         def test(total, fields):
    ##             return sum([total[field] for field in fields
    ##                         if field in total]) > nmin

    ##         for s in ys:
    ##             if test(tot, fields):
    ##                 break
    ##             # add more stats
    ##             for f in self.sumfields:
    ##                 tot[f] = (tot.setdefault(f, 0)
    ##                           + (s[f] * aux_weight if s[f] else 0))
    ##             tot['YWEIGHT'] = tot.setdefault('YWEIGHT', 0) + aux_weight

    return tot


class StatsTotal(object):

    minsample = (0, [])
    fields = ()

    @classmethod
    def weighted(cls, stats, currentyear, yweight):
        return weighted_total(stats, cls.fields, currentyear, yweight)

    @classmethod
    def normal(cls, stats):
        return total(stats, cls.fields)


class TotBatting(StatsTotal):

    minsample = (550, ['AB', 'BB', 'IBB', 'HBP', 'SH', 'SF'])
    fields = ('G', 'G_batting', 'AB', 'R', 'H', '2B', '3B', 'HR', 'RBI',
              'SB', 'CS', 'BB', 'SO', 'IBB', 'HBP', 'SH', 'SF', 'GIDP',
              'G_old')


class TotPitching(StatsTotal):

    minsample = (550, ['BFP'])
    fields = ('W', 'L', 'G', 'GS', 'CG', 'SHO', 'SV', 'IPouts', 'H', 'ER',
              'HR', 'BB', 'SO', 'IBB', 'WP', 'HBP', 'BK', 'BFP', 'GF', 'R',
              'SH', 'SF', 'GIDP')


class TotFielding(StatsTotal):

    minsample = (550, ['InnOuts'])
    fields = ('G', 'GS', 'InnOuts', 'PO', 'A', 'E', 'DP', 'PB', 'WP', 'SB',
              'CS')

    @classmethod
    def weighted(cls, stats, currentyear, yweight):
        m = ((stats['POS'] == 'P')
             + (stats['POS'] == 'C')
             + (stats['POS'] == '1B')
             + (stats['POS'] == '2B')
             + (stats['POS'] == '3B')
             + (stats['POS'] == 'SS')
             + (stats['POS'] == 'LF')
             + (stats['POS'] == 'CF')
             + (stats['POS'] == 'RF'))
        stats = stats[m]
        
        return weighted_total(stats, cls.fields, currentyear, yweight)


class TotAppearances(StatsTotal):

    minsample = (162, ['G_all'])
    fields = ('G_all', 'GS', 'G_batting', 'G_defense', 'G_p', 'G_c',
              'G_1b', 'G_2b', 'G_3b', 'G_ss', 'G_lf', 'G_cf', 'G_rf',
              'G_of', 'G_dh', 'G_ph', 'G_pr')


# class StatsTable(tuple):

#     # (minimum for statistical significance, [fields to sum])
#     minimum_opportunity = None

#     # fields to tally up
#     sumfields = ()

#     def __init__(self, rows):
#         super(StatsTable, self).__init__(null2zero(rows))

#     @property
#     def numseason(self):
#         return len(set([y['yearID'] for y in self]))

#     @property
#     def total(self):
#         tot = {}
#         for s in self:
#             for f in self.sumfields:
#                 tot[f] = tot.setdefault(f, 0) + (s[f] if s[f] else 0)
#         return tot

#     def total4rating(self, year, yweight=None, samplingmode='optimal'):
#         if len(self) == 0:
#             # no stats recorded, so no total
#             return {}

#         if samplingmode == 'career':
#             return self.total

#         def cmpkey(s):
#             return s['yearID']

#         # list all yearID values in order of summing.
#         ys_weighted = [y for y in self if y['yearID'] in yweight.keys()]
#         ys = sorted([y for y in self if y['yearID'] not in ys_weighted
#                      and y['yearID'] <= year], key=cmpkey, reverse=True)
#         ys.extend(sorted([y for y in self if y['yearID'] not in ys_weighted
#                           and y['yearID'] > year], key=cmpkey))

#         tot = {}
#         for s in ys_weighted:
#             for f in self.sumfields:
#                 yw = yweight[s['yearID']]
#                 tot[f] = (tot.setdefault(f, 0)
#                           + (s[f] * yw if s[f] else 0))
#             tot['YWEIGHT'] = tot.setdefault('YWEIGHT', 0) + yw

#         if self.minimum_opportunity is not None:
#             aux_weight = (min(yweight.values()) / 2. if len(yweight.values())
#                           else 1.)
#             nmin, fields = self.minimum_opportunity

#             def test(total, fields):
#                 return sum([total[field] for field in fields
#                             if field in total]) > nmin

#             for s in ys:
#                 if test(tot, fields):
#                     break
#                 # add more stats
#                 for f in self.sumfields:
#                     tot[f] = (tot.setdefault(f, 0)
#                               + (s[f] * aux_weight if s[f] else 0))
#                 tot['YWEIGHT'] = tot.setdefault('YWEIGHT', 0) + aux_weight

#         return tot




#!/usr/bin/env python2.7
from . import lahmandb


def find_teams_in_year(year, db=lahmandb):
    sql = "SELECT teamID, lgID, name FROM Teams WHERE yearID = %(year)s"
    c = db.cursor()
    c.execute(sql, {'year': year})
    return c.fetchall()


def find_players_in_team(year, team, db=lahmandb):
    params = {'team': str(team), 'year': int(year)}
    pids = []
    sql = ("SELECT playerID FROM Appearances"
           "  WHERE teamID = %(team)s AND yearID = %(year)s"
           "  ORDER BY G_p DESC, G_all DESC")
    c = db.cursor()
    c.execute(sql, params)
    return [o[0] for o in c.fetchall()]



#!/usr/bin/env python2.7
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
#from __future__ import unicode_literals
from math import exp, sqrt


def argsort(seq):
    """Emulate NumPy argsort"""
    return sorted(range(len(seq)), key=seq.__getitem__)


class require(object):

    def __init__(self, *args, **kwargs):
        self.requires = args
        self.retval = kwargs.get('retval', -1)

    def __call__(self, originalfunc):
        decoratorself = self
        def wrappee(s, *args, **kwargs):
            requires = decoratorself.requires
            if requires:
                for r in requires:
                    if len(s.stats[r]) == 0:
                        return decoratorself.retval
            return originalfunc(s, *args, **kwargs)
        return wrappee


def keeprange(fn):
    def _f(s, *args, **kwargs):
        r = fn(s, *args, **kwargs)
        if r == -1:
            return -1
        r = int(round(r))
        return r if 0 <= r <= 99 else (0 if r < 0 else 99)
    return _f


def div(a, b):
    return a / b if b != 0 else 0.


class RatingGenerator(object):

    def __init__(self, attr, stats):
        self.stats = stats
        self.attr = attr

    def rate(self):
        a = self.attr
        a['PO1'], a['PO2'] = self.find_positions()

        a['STA'] = self.rate_stamina()
        a['PCL'] = self.rate_pitchingclutch()
        a['K9'] = self.rate_k9()
        a['BB9'] = self.rate_bb9()
        a['H9'] = self.rate_h9()
        a['HR9'] = self.rate_hr9()

        a['RCT'] = self.rate_contact()
        a['LCT'] = self.rate_contact()
        a['RPW'] = self.rate_power()
        a['LPW'] = self.rate_power()
        a['BNT'] = self.rate_buntingability()
        a['DBN'] = self.rate_dragbunting()
        a['VIS'] = self.rate_platevision()
        a['DIS'] = self.rate_platediscipline()
        a['CLU'] = self.rate_clutch()
        a['DUR'] = self.rate_durability()
        a['SPD'] = self.rate_speed(pos=a['PO1'])
        a['BAB'] = self.rate_brability()
        a['BAG'] = self.rate_braggressiveness()

        return a

    def rate_stamina(self, *args, **kwargs): return -1
    def rate_pitchingclutch(self, *args, **kwargs): return -1
    def rate_k9(self, *args, **kwargs): return -1
    def rate_bb9(self, *args, **kwargs): return -1
    def rate_h9(self, *args, **kwargs): return -1
    def rate_hr9(self, *args, **kwargs): return -1
    def rate_contact(self, *args, **kwargs): return -1
    def rate_power(self, *args, **kwargs): return -1
    def rate_buntingability(self, *args, **kwargs): return -1
    def rate_dragbunting(self, *args, **kwargs): return -1
    def rate_platevision(self, *args, **kwargs): return -1
    def rate_platediscipline(self, *args, **kwargs): return -1
    def rate_clutch(self, *args, **kwargs): return -1
    def rate_durability(self, *args, **kwargs): return -1
    def rate_speed(self, *args, **kwargs): return -1
    def rate_brability(self, *args, **kwargs): return -1
    def rate_braggressiveness(self, *args, **kwargs): return -1
    def find_positions(self): return ('', '')


class BasicRatingGenerator(RatingGenerator):

    def rate(self):
        a = super(BasicRatingGenerator, self).rate()
        a['RCT'] = self.rate_contact(bat=a['B'], vs='R', displ=a['DIS'])
        a['LCT'] = self.rate_contact(bat=a['B'], vs='L', displ=a['DIS'])
        a['RPW'] = self.rate_power(bat=a['B'], vs='R')
        a['LPW'] = self.rate_power(bat=a['B'], vs='L')
        return a

    @keeprange
    @require('pitching')
    def rate_stamina(self, *args, **kwargs):
        s = self.stats['pitching']
        if self.attr['PO1'] == 'SP':
            x = (s['IPouts'] / 3. / s['GS'])
            r = (100. - 60.) / (7.7 - 4.3) * (x - 4.3) + 60.
            r = max(r, 60.)
        elif self.attr['PO1'] in ('RP', 'CP'):
            x = (s['IPouts'] / 3. / s['G'])
            r = (50. - 15.) / (2.0 - 0.5) * (x - 1.0) + 28.
            r = max(r, 15.)
        else:
            r = -1
        return r

    @keeprange
    @require('pitching')
    def rate_pitchingclutch(self, *args, **kwargs):
        return 54

    @keeprange
    @require('pitching')
    def rate_k9(self, *args, **kwargs):
        s = self.stats['pitching']
        so, out = s.get('SO', -1), s.get('IPouts', -1)
        if not (out > 0):
            return -1
        k9 = 27. * so / out
        return (k9 + 0.6519) / 0.1139

    @keeprange
    @require('pitching')
    def rate_bb9(self, *args, **kwargs):
        s = self.stats['pitching']
        bb, tbf = s.get('BB', -1), s.get('BFP', -1)
        if not (tbf > 0):
            return -1
        rr = bb / tbf
        return (rr - 0.1309) / -0.0007

    @keeprange
    @require('pitching')
    def rate_h9(self, *args, **kwargs):
        s = self.stats['pitching']
        h, ab  = s['H'], s['BFP'] - s['BB'] -s['HBP'] -s['SF'] -s['SH']
        if not (ab > 0):
            return -1
        ba = h / ab
        return (ba - 0.3477) / -0.0014

    @keeprange
    @require('pitching')
    def rate_hr9(self, *args, **kwargs):
        s = self.stats['pitching']
        hr, h  = s['HR'], s['H']
        if not (h > 0):
            return -1
        x = 1. * hr / h
        r = (x - 0.11) * (100. - 20.) / (0.05 - 0.17) + 60.
        r = max(r, 20.)
        return r

    @keeprange
    @require('batting')
    def rate_contact(self, bat='R', vs='R', displ=40, psig=0., *args, **kwargs):
        s = self.stats['batting']
        
        h, ab = s['H'], s['AB']
        if not (ab > 0):
            return -1
        # generic ratio for (AB vs LHP) / (AB vs RHP)
        r = .50786 if bat == 'R' else 0.21818

        # platoon advantage: [(OBA vs LHP) - (OBA vs RHP)] / (OBA total)
        # for RHB.
        pltp = {'R': 0.0514, 'L': -0.0833, 'S': 0.0000}[bat]

        # std dev
        spltp = {'R': 0.0420, 'L': 0.0458, 'S': 0.0640}[bat]
        pltp += spltp * psig if bat == 'R' else -spltp * psig
        
        ba = (1.*h / ab * (1 + r * (1 - pltp)) / (1 + r) if vs == 'R'
              else (1.*h / ab * (r + 1 + pltp)) / (1 + r))
        return ((1. * ba - 0.1838 - 1.965e-4 * displ)
                / (1.034e-3 + 1.508e-6 * displ))

    @keeprange
    @require('batting')
    def rate_power(self, bat='R', vs='R', psig=0., *args, **kwargs):
        s = self.stats['batting']

        h, hr = s['H'], s['HR'] 
        if not (h > 0):
            return -1

        # generic ratio for (AB vs LHP) / (AB vs RHP)
        r = .50786 if bat == 'R' else 0.21818

        # platoon advantage: [(OBA vs LHP) - (OBA vs RHP)] / (OBA total)
        # for RHB.
        pltp = {'R': 0.0095, 'L': -0.0520, 'S': 0.000}[bat]
        
        # std dev
        spltp = {'R': 0.0420, 'L': 0.0458, 'S': 0.0640}[bat]
        pltp += spltp * psig if bat == 'R' else -spltp * psig

        hrph = (1.*hr / h * (1 + r * (1 - pltp)) / (1 + r) if vs == 'R'
                else (1.*hr / h * (r + 1 + pltp)) / (1 + r))
        return (hrph + 0.06375) / 3.350e-3

    @keeprange
    @require('batting')
    def rate_buntingability(self, *args, **kwargs):
        s = self.stats['batting']
        sh, pa = s['SH'], s['AB'] + s['BB'] + s['SH'] + s['SF'] + s['HBP']
        shpa = (div(1. * sh, pa) / 12. if self.attr['PO1'] in ('SP','RP','CP')
                else 1. * div(sh, pa))
        if not (pa > 0):
            return -1
        return 93. * (1. - exp(-80. * (shpa)))

    @keeprange
    @require('batting')
    def rate_dragbunting(self, *args, **kwargs):
        s = self.stats['batting']
        sh, pa = s['SH'], s['AB'] + s['BB'] + s['SH'] + s['SF'] + s['HBP']
        if not (pa > 0):
            return -1
        return 96. * (1. - exp(-60. * ((1. * sh / pa) - 0.002)))

    @keeprange
    @require('batting')
    def rate_platevision(self, *args, **kwargs):
        s = self.stats['batting']
        so, ab = s['SO'], s['AB']
        if not (ab > 0):
            return -1
        return - ((1. * so / ab) - 0.2884) / 1.6732e-3

    @keeprange
    @require('batting')
    def rate_platediscipline(self, *args, **kwargs):
        s = self.stats['batting']
        bb, pa = s['BB'], s['AB'] + s['BB'] + s['HBP'] + s['SH'] + s['SF']
        if not (pa > 0):
            return -1
        return (771.1 * (bb / pa))

    @keeprange
    @require('batting')
    def rate_clutch(self, *args, **kwargs):
        return 70.

    @keeprange
    @require('batting', 'fielding', 'appearances')
    def rate_durability(self, *args, **kwargs):
        s = self.stats['appearances']
        g_all, gs, gp = s['G_all'], s['GS'], s['G_p']
        yrtot = s['YR_TOT']

        if self.attr['PO1'] == 'SP':
            return 55. * gs / yrtot / 36. + 45.
        elif self.attr['PO1'] == 'RP':
            return 70. * gp / yrtot / 80. + 30.
        elif self.attr['PO1'] == 'CP':
            return 55. * gp / yrtot / 50. + 45.

        s = self.stats['fielding']
        innouts = s['InnOuts']

        s = self.stats['batting']
        pa = sum([s[k] for k in ['AB', 'BB', 'IBB', 'HBP', 'SH', 'SF']])
        paneed = g_all * 3.1
        
        r = 1. * (innouts / 3.) / (g_all * 9.)
        
        if 1. * gs / g_all > 0.8:
            # starter...
            gsf = (innouts / 3.) / (g_all * 9.)
            r1 = (45. if gsf < .75
                  else ((gsf - .75)*(95. - 45.)/(.9 - .75) + 45 if gsf < .9
                        else (gsf - .95)*(99. - 95.)/(1. - .95) + 95))
            r2 = 70. * g_all / (yrtot * 162.) + 30.
            r = min(r1, r2)
        else:
            # bench player...
            r = 70. * g_all / (yrtot * 162.) + 30.
        return r

    @keeprange
    @require('batting', 'fielding')
    def rate_speed(self, *args, **kwargs):
        pos = kwargs.get('pos', None)
        s = self.stats['batting']
        sb, cs = 1. * s['SB'], 1. * s['CS']
        h, h2, h3, hr = 1. * s['H'], 1. * s['2B'], 1. * s['3B'], 1. * s['HR']
        ab, so, bb = 1.*s['AB'], 1.*s['SO'], 1.*s['BB']
        r, hbp, gidp = 1.*s['R'], 1.*s['HBP'], 1.*s['GIDP']

        s = self.stats['fielding']
        po, a, g = 1. * s['PO'], 1. * s['A'], 1. * s['G']

        f1 = (sb + 3.0) / (sb + cs + 7.0)
        f1 = (f1 - 0.4) * 20.
        f2 = div((sb + cs) , ((h - h2 - h3 - hr) + bb + hbp))
        f2 = sqrt(f2) / 0.07
        f3 = div(h3 , (ab - hr - so))
        f3 = f3 / 0.0016
        f4 = div((r - hr) , (h + bb + hbp - hr))
        f4 = (f4 - 0.1) * 25.
        f5 = div(gidp , (ab - hr - so))
        f5 = (0.063 - f5) / 0.007
        f6 = {None: 0, '': 0,
              'SP': 0., 'RP': 0., 'CP': 0.,
              'C': 1.,
              '1B': 2.,
              '2B': (((po + a) / g) / 4.8) * 6.,
              '3B': (((po + a) / g) / 2.65) * 4.,
              'SS': (((po + a) / g) / 4.6) * 7.,
              'LF': (((po + a) / g) / 2.0) * 6.,
              'CF': (((po + a) / g) / 2.0) * 6.,
              'RF': (((po + a) / g) / 2.0) * 6.,
              'OF': (((po + a) / g) / 2.0) * 6.}[pos]

        ss = [max(min(float(f), 10.), 0.) for f in [f1, f2, f3, f4, f5, f6]]
        ss = float(sum(ss) / len(ss))
        r = ((ss - 1.531) / 1.498e-3)**(1./1.822) if ss > 1.531 else 0.
        return r

    @keeprange
    @require('batting')
    def rate_brability(self, *args, **kwargs):
        s = self.stats['batting']
        sb, cs = s['SB'], s['CS']
        if not (sb + cs > 0):
            return -1
        sbp = 1. * sb / (sb + cs)
        return ((sbp * 100 - 52.3) / 0.3525)

    @keeprange
    @require('batting')
    def rate_braggressiveness(self, *args, **kwargs):
        s = self.stats['batting']
        sb, cs, h, bb, hbp = s['SB'], s['CS'], s['H'], s['BB'], s['HBP']
        if not (h + bb + hbp > 0):
            return -1
        sbr = 1. * (sb + cs) / (h + bb + hbp)
        return ((sbr + .02169) / 0.003397)

    @require('appearances', retval=('', ''))
    def find_positions(self):
        stats = self.stats

        s = stats['appearances']
        posstr = ['P', 'C', '1B', '2B', '3B', 'SS', 'LF', 'CF', 'RF']
        gs = [s['G_p'], s['G_c'], s['G_1b'],
              s['G_2b'], s['G_3b'], s['G_ss'],
              s['G_lf'], s['G_cf'], s['G_rf']]
        if sum(gs) == 0:
            # likely a DH; need to decide from defensive spectrum
            return ('LF', '1B')

        idxsorted = argsort(gs)[::-1]
        pos1 = idxsorted[0]

        if pos1 == 0:
            # this is a pitcher, so choose one of SP, RP, and CP:
            p = stats['pitching']
            if len(p) == 0:
                return ('', '')
            if (1. * p['GS']) / p['G'] > .5:
                return ('SP', '')
            else:
                return ('CP' if p['SV'] > p['G'] * .3 else 'RP', '')

        if gs[idxsorted[1]] == 0:
            return (posstr[pos1], '')
        if gs[idxsorted[2]] == 0:
            return (posstr[pos1], posstr[idxsorted[1]])

        # below, we are dealing with a case in which more than two
        # positions have been played by this player.
        nif = sum(gs[2:6]) if len([1 for x in gs[2:6] if x > 0]) == 4 else 0
        nof = sum(gs[6:9]) if len([1 for x in gs[6:9] if x > 0]) == 3 else 0
        if nif > 0 and nif >= nof:
            return (posstr[pos1], 'IF')
        if nof > 0 and nof > nif:
            return (posstr[pos1], 'OF')
        if pos1 not in [3, 5] and gs[3] > 0 and gs[5] > 0:
            return (posstr[pos1], '2B/SS')
        if pos1 not in [2, 4] and gs[2] > 0 and gs[4] > 0:
            return (posstr[pos1], '1B/3B')
        if pos1 not in [1, 2] and gs[1] > 0 and gs[2] > 0:
            return (posstr[pos1], 'C/1B')
        if pos1 not in [1, 4] and gs[1] > 0 and gs[4] > 0:
            return (posstr[pos1], 'C/3B')
        return (posstr[pos1], posstr[idxsorted[1]])



#!/usr/bin/env python2.7
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
#from __future__ import unicode_literals
from theshowutil.playerdata import TotPitching
from theshowutil.playerdata import TotBatting
from theshowutil.playerdata import TotFielding
from theshowutil.playerdata import TotAppearances


def outs2inn(outs):
    outs = int(round(outs))
    return ('{}.{}'.format(outs // 3, outs % 3) if outs % 3 > 0
            else '{}  '.format(outs // 3))


class ViewStats(object):

    header = ''
    fmt = ''

    def __init__(self, playerdata, stats):
        self.playerdata = playerdata
        self.stats = stats

    def compute_total(self):
        return {}

    def compute_per_season(self, total):
        return {}

    def render(self):
        print(self.header)
        print('-' * len(self.header))
        stats = self.stats.filled(0.)
        for eachstats in stats:
            o = self.rendereach(eachstats)
            print(self.fmt.format(**o))

        tot = self.compute_total()
        n162 = self.compute_per_season(tot)

        print('-' * len(self.header))

        self.rendertotal(tot)
        yrtot = len(set(o['yearID'] for o in self.stats))
        tot['INFO'] = '{:2} Season Total'.format(yrtot)
        print(self.fmt.format(**tot))
        
        self.rendertotal(n162)
        n162['INFO'] = '162 Game Ave.'
        print(self.fmt.format(**n162))

    def rendereach(self, eachstats):
        return {}

    def rendertotal(self, total):
        return {}


class ViewPitchingStats(ViewStats):

    header = ('  YR AGE  TM LG   W   L    ERA   G  GS  CG  SV    IP'
              '      H    R   ER   HR   BB   SO')
    fmt = ('{INFO:>15} {W:3.0f} {L:3.0f}'
           ' {ERA:6.2f} {G:3.0f} {GS:3.0f} {CG:3.0f} {SV:3.0f} {IP:>7}'
           ' {H:4.0f} {R:4.0f} {ER:4.0f} {HR:4.0f} {BB:4.0f} {SO:4.0f}')

    def compute_total(self):
        return TotPitching.normal(self.stats)

    def compute_per_season(self, total):
        # compute 162 G ave (following baseball-reference.com convention)
        n162 = {}
        for col in TotPitching.fields:
            n162[col] = 68. * total[col] / (total['G'] + total['GS'])
        return n162

    def rendereach(self, eachstats):
        s = eachstats
        o = {}
        for n in s.dtype.names:
            o[n] = s[n]
        o['age'] = self.playerdata.age(o['yearID'])
        o['INFO'] = ('{yearID:4d} {age:3d} {teamID:3s} {lgID:2s}'
                     .format(**o))
        o['ERA'] = 27. * o['ER'] / o['IPouts']
        o['IP'] = outs2inn(o['IPouts'])
        return o

    def rendertotal(self, total):
        o = total
        o['ERA'] = 27. * o['ER'] / o['IPouts']
        o['IP'] = outs2inn(o['IPouts'])
        return o


def fmtrate(r):
    return (' ---' if r is None
            else (('{:.3f}'.format(r))[1:] if r < 1
                  else ('{:.3f}'.format(r))[:4]))

def rate(x, y):
    return 1. * x / y if abs(y) > 0 else None


class ViewBattingStats(ViewStats):

    header = ('  YR AGE  TM LG'
              '    G    AB    R    H  2B  3B  HR  RBI'
              '  SB  CS   BB   SO'
              '   BA  OBP  SLG  OPS')
    fmt = ('{INFO:>15} {G:4.0f} {AB:5.0f} {R:4.0f} {H:4.0f} {2B:3.0f} {3B:3.0f}'
           ' {HR:3.0f} {RBI:4.0f} {SB:3.0f} {CS:3.0f} {BB:4.0f} {SO:4.0f}'
           ' {BA:4s} {OBP:4s} {SLG:4s} {OPS:4s}')

    def compute_total(self):
        return TotBatting.normal(self.stats)

    def compute_per_season(self, total):
        # compute 162 G ave (following baseball-reference.com convention)
        n162 = {}
        for col in TotBatting.fields:
            n162[col] = 162. * total[col] / total['G']
        return n162

    def rendereach(self, eachstats):
        s = eachstats
        o = {}
        for n in s.dtype.names:
            o[n] = s[n]
        o['age'] = self.playerdata.age(o['yearID'])
        o['INFO'] = ('{yearID:4d} {age:3d} {teamID:3s} {lgID:2s}'.format(**o))
        o['BA'] = fmtrate(rate(o['H'], o['AB']))
        obp = rate(o['H'] + o['BB'] + o['HBP'],
                   o['AB'] + o['BB'] + o['HBP'] + o['SF'])
        o['OBP'] = fmtrate(obp)
        slg = rate(o['H'] + 2. * o['2B'] + 3. * o['3B'] + 4. * o['HR'],
                   o['AB'])
        o['SLG'] = fmtrate(slg)
        o['OPS'] = fmtrate(None if None in (obp, slg) else obp + slg)
        return o

    def rendertotal(self, total):
        o = total
        o['BA'] = fmtrate(rate(o['H'], o['AB']))
        obp = rate(o['H'] + o['BB'] + o['HBP'],
                   o['AB'] + o['BB'] + o['HBP'] + o['SF'])
        o['OBP'] = fmtrate(obp)
        slg = rate(o['H'] + 2. * o['2B'] + 3. * o['3B'] + 4. * o['HR'],
                   o['AB'])
        o['SLG'] = fmtrate(slg)
        o['OPS'] = fmtrate(None if None in (obp, slg) else obp + slg)
        return o


class ViewFieldingStats(ViewStats):

    header = ('  YR AGE  TM LG POS    G   GS   INN      PO    A   E   DP'
              '  PB  WP   SB   CS')
    fmt = ('{INFO:>19} {G:4.0f} {GS:4.0f} {INN:>7s} {PO:5.0f}'
           ' {A:4.0f} {E:3.0f} {DP:4.0f} {PB:3.0f} {WP:3.0f} {SB:4.0f}'
           ' {CS:4.0f}')

    def compute_total(self):
        return TotFielding.normal(self.stats)

    def compute_per_season(self, total):
        # compute 162 G ave (following baseball-reference.com convention)
        n162 = {}
        for col in TotFielding.fields:
            n162[col] = 162. * total[col] / total['G']
        return n162

    def rendereach(self, eachstats):
        s = eachstats
        o = {}
        for n in s.dtype.names:
            o[n] = s[n]
        o['age'] = self.playerdata.age(o['yearID'])
        o['INFO'] = ('{yearID:4d} {age:3d} {teamID:3s}'
                     ' {lgID:2s} {POS:3s}'.format(**o))
        o['INN'] = outs2inn(o['InnOuts'])
        return o

    def rendertotal(self, total):
        o = total
        o['INN'] = outs2inn(o['InnOuts'])
        return o


class ViewAppearancesStats(ViewStats):

    header = ('  YR AGE  TM LG'
              '    G   GS  BAT  DEF    P    C   1B   2B   3B   SS'
              '   LF   CF   RF   OF   DH')
    fmt = ('{INFO:>15}'
           ' {G_all:4.0f} {GS:4.0f} {G_batting:4.0f} {G_defense:4.0f} {G_p:4.0f} {G_c:4.0f} {G_1b:4.0f} {G_2b:4.0f} {G_3b:4.0f}'
           ' {G_ss:4.0f} {G_lf:4.0f} {G_cf:4.0f} {G_rf:4.0f} {G_of:4.0f} {G_dh:4.0f}')

    def compute_total(self):
        return TotAppearances.normal(self.stats)

    def compute_per_season(self, total):
        # compute 162 G ave (following baseball-reference.com convention)
        n162 = {}
        for col in TotAppearances.fields:
            n162[col] = 162. * total[col] / total['G_all']
        return n162

    def rendereach(self, eachstats):
        s = eachstats
        o = {}
        for n in s.dtype.names:
            o[n] = s[n]
        o['age'] = self.playerdata.age(o['yearID'])
        o['INFO'] = ('{yearID:4d} {age:3d} {teamID:3s} {lgID:2s}'.format(**o))
        return o

    def rendertotal(self, total):
        return total


def showstats(pd):
    if len(pd.pitching):
        print('PITCHING\n')
        v = ViewPitchingStats(pd, pd.pitching)
        v.render()
        print()
    if len(pd.batting):
        print('BATTING\n')
        v = ViewBattingStats(pd, pd.batting)
        v.render()
        print()
    if len(pd.fielding):
        print('FIELDING\n')
        v = ViewFieldingStats(pd, pd.fielding)
        v.render()
        print()
    if len(pd.appearances):
        print('APPEARANCES\n')
        v = ViewAppearancesStats(pd, pd.appearances)
        v.render()


def showratings(attr):
    print('FNM        LNM           HT     WT  PO1  PO2       T  B    USB')
    print('{FNM:<10s} {LNM:<10s}   {HT:5s}  {WT:3d}   {PO1:4s} {PO2:4s}'
          '     {T:1s}  {B:1s}      {USB:1s}'
          .format(**attr))
    print()

    if attr['PO1'] in ['SP', 'RP', 'CP']:
        print('STA  PCL   H9  HR9   K9  BB9')
        print(' {STA:2d}   {PCL:2d}   {H9:2d}   {HR9:2d}   {K9:2d}   {BB9:2d}'
              .format(**attr))
        print()

        h, s = '', ''
        for i in range(1, 6):
            key = 'P{:d}'.format(i)
            if len(attr[key]) == 0:
                continue
            h += '  P{i:d} P{i:d}S P{i:d}C P{i:d}B '.format(i=i)
            s += ('{:4s}  {:2d}  {:2d}  {:2d} '
                  .format(attr[key], attr[key + 'S'], attr[key +'C'],
                          attr[key + 'B']))
        if h and s:
            print(h)
            print(s)
            print()

    print('RCT  LCT  RPW  LPW  BNT  DBN  VIS  DIS'
          '  CLU  DUR  SPD  AMS  AMA  REA  FLD  BLK  BAB  BAG')
    print(' {RCT:2d}   {LCT:2d}   {RPW:2d}   {LPW:2d}   {BNT:2d}   {DBN:2d}'
          '   {VIS:2d}   {DIS:2d}   {CLU:2d}   {DUR:2d}   {SPD:2d}   {AMS:2d}'
          '   {AMA:2d}   {REA:2d}   {FLD:2d}   {BLK:2d}   {BAB:2d}   {BAG:2d}'
          .format(**attr))



#!/usr/bin/env python2.7
"""
MLB the Show Player Rating Utility
"""
import os
from setuptools import setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(name='theshowutil',
      version='0.1.1',
      author='Taro Sato',
      author_email='okomestudio@gmail.com',
      url='https://github.com/nomo17k/theshowutil',
      description='MLB the Show player rating utility',
      long_description=read('README'),
      license='BSD',
      packages=['theshowutil'],
      package_dir={'theshowutil': 'theshowutil'},
      scripts=['bin/theshowrater',
               'bin/theshowfinder'],
      classifiers=["Development Status :: 3 - Alpha"]
      )



#!/usr/bin/env python2.7
"""Lahman database driver module

This module provides a singleton interface to the Lahman database.
The instantiation of database connection is done lazily to prevent
openning a connection to be never used, since the web-based
application has its own flow of using the database and would not use
this module.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
#from __future__ import unicode_literals
import os
import MySQLdb
import numpy as np


DATABASE = 'lahman591'
DEFAULT_FILE = '/'.join([os.environ.get('HOME', '~'), ".my.cnf"])


_db = None


def connect():
    global _db
    if _db:
        return _db
    _db = MySQLdb.connect(db=DATABASE,
                          read_default_file=DEFAULT_FILE)
    return _db


def cursor(cursorclass=None):
    global _db
    _db = _db if _db else connect()
    if cursorclass:
        return _db.cursor(cursorclass)
    return _db.cursor()


class LahmanReader(object):
    """Lahman database conversion utility"""

    dtype = []

    @classmethod
    def fromiter(cls, iter):
        d = np.array([], dtype=cls.dtype)
        mask = []
        for size, o in enumerate(iter):
            d.resize(size + 1)
            ms = np.zeros((len(o),), dtype=bool)
            for i, x in enumerate(o):
                dt = cls.dtype[i][1][0]
                ms[i] = (x is None)
                x = (x if x is not None
                     else (-99999 if dt == 'i'
                           else (-99999. if dt == 'f'
                                 else '')))
                d[-1][i] = x
            mask.append(tuple(ms))
        d = np.ma.array(d, mask=mask)
        return d


class Master(LahmanReader):

    dtype = [('lahmanID', 'i'),
             ('playerID', 'a9'),
             ('managerID', 'a10'),
             ('hofID', 'a10'),
             ('birthYear', 'i'),
             ('birthMonth', 'i'),
             ('birthDay', 'i'),
             ('birthCountry', 'a24'),
             ('birthState', 'a2'),
             ('birthCity', 'a31'),
             ('deathYear', 'i'),
             ('deathMonth', 'i'),
             ('deathDay', 'i'),
             ('deathCountry', 'a34'),
             ('deathState', 'a2'),
             ('deathCity', 'a21'),
             ('nameFirst', 'a12'),
             ('nameLast', 'a14'),
             ('nameNote', 'a80'),
             ('nameGiven', 'a43'),
             ('nameNick', 'a48'),
             ('weight', 'f'),
             ('height', 'f'),
             ('bats', 'a1'),
             ('throws', 'a1'),
             ('debut', 'a18'),
             ('finalGame', 'a18'),
             ('college', 'a40'),
             ('lahman40ID', 'a9'),
             ('lahman45ID', 'a9'),
             ('retroID', 'a8'),
             ('holtzID', 'a9'),
             ('bbrefID', 'a9')]


class Batting(LahmanReader):
    
    dtype = [("playerID", 'a9'),
             ("yearID", 'i'),
             ("stint", 'i'),
             ("teamID", 'a3'),
             ("lgID", 'a2'),
             ("G", 'i'),
             ("G_batting", 'i'),
             ("AB", 'i'),
             ("R", 'i'),
             ("H", 'i'),
             ("2B", 'i'),
             ("3B", 'i'),
             ("HR", 'i'),
             ("RBI", 'i'),
             ("SB", 'i'),
             ("CS", 'i'),
             ("BB", 'i'),
             ("SO", 'i'),
             ("IBB", 'i'),
             ("HBP", 'i'),
             ("SH", 'i'),
             ("SF", 'i'),
             ("GIDP", 'i'),
             ("G_old", 'i')]


class Pitching(LahmanReader):

    dtype = [("playerID", 'a9'),
             ("yearID", 'i'),
             ("stint", 'i'),
             ("teamID", 'a3'),
             ("lgID", 'a2'),
             ("W", 'i'),
             ("L", 'i'),
             ("G", 'i'),
             ("GS", 'i'),
             ("CG", 'i'),
             ("SHO", 'i'),
             ("SV", 'i'),
             ("IPouts", 'i'),
             ("H", 'i'),
             ("ER", 'i'),
             ("HR", 'i'),
             ("BB", 'i'),
             ("SO", 'i'),
             ("BAOpp", 'f'),
             ("ERA", 'f'),
             ("IBB", 'i'),
             ("WP", 'i'),
             ("HBP", 'i'),
             ("BK", 'i'),
             ("BFP", 'i'),
             ("GF", 'i'),
             ("R", 'i'),
             ("SH", 'i'),
             ("SF", 'i'),
             ("GIDP", 'i')]


class Fielding(LahmanReader):

    dtype = [("playerID", 'a9'),
             ("yearID", 'i'),
             ("stint", 'i'),
             ("teamID", 'a3'),
             ("lgID", 'a2'),
             ("POS", 'a2'),
             ("G", 'i'),
             ("GS", 'i'),
             ("InnOuts", 'i'),
             ("PO", 'i'),
             ("A", 'i'),
             ("E", 'i'),
             ("DP", 'i'),
             ("PB", 'i'),
             ("WP", 'i'),
             ("SB", 'i'),
             ("CS", 'i'),
             ("ZR", 'f')]


class Appearances(LahmanReader):

    dtype = [("yearID", 'i'),
             ("teamID", 'a3'),
             ("lgID", 'a2'),
             ("playerID", 'a9'),
             #("UNKNOWN", 'i'),
             ("G_all", 'i'),
             ("GS", 'i'),
             ("G_batting", 'i'),
             ("G_defense", 'i'),
             ("G_p", 'i'),
             ("G_c", 'i'),
             ("G_1b", 'i'),
             ("G_2b", 'i'),
             ("G_3b", 'i'),
             ("G_ss", 'i'),
             ("G_lf", 'i'),
             ("G_cf", 'i'),
             ("G_rf", 'i'),
             ("G_of", 'i'),
             ("G_dh", 'i'),
             ("G_ph", 'i'),
             ("G_pr", 'i')]
